
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#ifndef _MOOF_MIPPLETON_HH_
#define _MOOF_MIPPLETON_HH_

/**
 * @file Mippleton.hh
 * Related to singletons, a mippleton is an object which can be obtained
 * globally using a unique name.  Unlike singletons, there can be multiple
 * mippletons per class, each with a different name or identifier.  Mippletons
 * are create automatically when they are first requested (retained) and deleted
 * after the last interested code releases its hold on the object.
 */

#include <string>

#include <boost/shared_ptr.hpp>

#include <Moof/Hash.hh>


namespace Mf {


template <class T>
class Mippleton
{
	typedef std::pair<unsigned,T*>							PtrValue;
	typedef stlplus::hash<std::string,PtrValue,getHash>		PtrMap;

	static PtrMap	ptrs_;
	std::string		name_;

	inline static T* retain(const std::string& name)
	{
		typename PtrMap::iterator it = ptrs_.find(name);

		if (it != ptrs_.end())
		{
			++((*it).second.first);
			return (*it).second.second;
		}
		else
		{
			T* newObj = new T(name);
			ptrs_.insert(std::make_pair(name, std::make_pair(1, newObj)));
			return newObj; 
		}
	}

	inline static void release(T* obj)
	{
		releaseByName(obj->name_);
	}

	inline static void releaseByName(const std::string& name)
	{
		typename PtrMap::iterator it;

		if ((it = ptrs_.find(name)) != ptrs_.end() && --(*it).second.first == 0)
		{
			delete (*it).second.second;
			ptrs_.erase((*it).first);
		}
	}

public:
	explicit Mippleton(const std::string& name) :
		name_(name) {}

	inline const std::string& getName() const
	{
		return name_;
	}

	inline static boost::shared_ptr<T> getInstance(const std::string& name)
	{
		return boost::shared_ptr<T>(retain(name), &release);
	}
};

template <class T>
stlplus::hash< std::string,std::pair<unsigned,T*>,getHash >
	Mippleton<T>::ptrs_;


} // namespace Mf

#endif // _MOOF_MIPPLETON_HH_

/** vim: set ts=4 sw=4 tw=80: *************************************************/

